// $Id: TCP_Contents.cc,v 1.15 2005/12/20 03:36:59 vern Exp $
//
// Copyright (c) 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#include "File.h"
#include "TCP.h"

namespace {
	const bool DEBUG_tcp_contents = false;
	const bool DEBUG_tcp_match_undelivered = false;
	const bool DEBUG_tcp_connection_close = false;
}

// Maximum amount of data that might plausibly be sent in an initial
// flight (prior to receiving any acks).  Used to determine whether we
// must not be seeing our peer's acks.
const int max_initial_window = 4096;

// If we're not seeing our peer's acks, maximum volume of data above
// a sequence hole that we'll tolerate before assuming that there's
// been a packet drop and we should give up on tracking a connection.
const int max_above_hole_without_any_acks = 4096;

TCP_Reassembler::~TCP_Reassembler()
	{
	if ( record_contents_file )
		{ // Record any undelivered data.
		if ( blocks && last_reassem_seq < last_block->upper )
			RecordToSeq(last_reassem_seq, last_block->upper,
					record_contents_file);
		Unref(record_contents_file);
		}
	}

void TCP_Reassembler::SizeBufferedData(int& waiting_on_hole,
					int& waiting_on_ack) const
	{
	waiting_on_hole = waiting_on_ack = 0;
	for ( DataBlock* b = blocks; b; b = b->next )
		{
		if ( b->seq <= last_reassem_seq )
			// We must have delivered this block, but
			// haven't yet trimmed it.
			waiting_on_ack += b->Size();
		else
			waiting_on_hole += b->Size();
		}
	}

void TCP_Reassembler::SetContentsFile(BroFile* f)
	{
	if ( ! f->IsOpen() )
		{
		run_time("no such file \"%s\"", f->Name());
		return;
		}

	if ( record_contents_file )
		// We were already recording, no need to catch up.
		Unref(record_contents_file);
	else
		{
		if ( blocks )
			RecordToSeq(blocks->seq, last_reassem_seq, f);
		}

	Ref(f);
	record_contents_file = f;
	}

void TCP_Reassembler::SkipToSeq(int seq)
	{
	TrimToSeq(seq);
	}

void TCP_Reassembler::Undelivered(int up_to_seq)
	{
	TCP_Endpoint* endpoint = contents_processor->Endpoint();
	TCP_Endpoint* peer = endpoint->peer;

	if ( up_to_seq <= 2 && contents_processor->Conn()->IsPartial() )
		// Since it was a partial connection, we faked up its
		// initial sequence numbers as though we'd seen a SYN.
		// We've now received the first ack and are getting a
		// complaint that either that data is missing (if
		// up_to_seq is 1), or one octet beyond it is missing
		// (if up_to_seq is 2).  The latter can occur when the
		// first packet we saw instantiating the partial connection
		// was a keep-alive.  So, in either case, just ignore it.
		return;

#if 0
	if ( endpoint->FIN_cnt > 0 )
		{
		// Make sure we're not worrying about undelivered
		// FIN control octets!
		int FIN_seq = endpoint->FIN_seq - endpoint->start_seq;
		if ( up_to_seq >= FIN_seq )
			up_to_seq = FIN_seq - 1;
		}
#endif

	if ( DEBUG_tcp_contents )
		{
		DEBUG_MSG("%.6f Undelivered: up_to_seq=%d, last_reassm=%d, "
		          "endp: FIN_cnt=%d, RST_cnt=%d, "
		          "peer: FIN_cnt=%d, RST_cnt=%d\n",
		          network_time, up_to_seq, last_reassem_seq,
		          endpoint->FIN_cnt, endpoint->RST_cnt,
		          peer->FIN_cnt, peer->RST_cnt);
		}

	if ( up_to_seq <= last_reassem_seq )
		return;

	if ( last_reassem_seq == 1 &&
	     (endpoint->FIN_cnt > 0 || endpoint->RST_cnt > 0 ||
	      peer->FIN_cnt > 0 || peer->RST_cnt > 0) )
		{
		// We could be running on a SYN/FIN/RST-filtered trace - don't
		// complain about data missing at the end of the connection.
		//
		// ### However, note that the preceding test is not a precise
		// one for filtered traces, and may fail, for example, when
		// the SYN packet carries data.
		//
		// Note, this check will confuse the EOF checker (and cause a
		// missing FIN in the rewritten trace) when the content gap
		// in the middle is discovered only after the FIN packet.

		// Skip the undelivered part without reporting to the endpoint
		}
	else
		{
		// Otherwise inform the endpoint about the gap.
		contents_processor->Undelivered(last_reassem_seq,
						up_to_seq - last_reassem_seq);
		}

	// We should record and match undelivered even if we are skipping
	// content gaps between SYN and FIN, because FIN may carry some data.

	if ( record_contents_file )
		RecordToSeq(last_reassem_seq, up_to_seq, record_contents_file);

	if ( tcp_match_undelivered )
		MatchUndelivered(up_to_seq);

	// But we need to re-adjust last_reassem_seq in either case.
	last_reassem_seq = up_to_seq;	// we've done our best ...
	}

void TCP_Reassembler::MatchUndelivered(int up_to_seq)
	{
	if ( ! blocks || ! rule_matcher )
		return;

	ASSERT(last_block);
	if ( up_to_seq == -1 )
		up_to_seq = last_block->upper;

	// ### Note: the original code did not check whether blocks have
	// already been delivered, but not ACK'ed, and therefore still
	// must be kept in the reassember.

	// We are to match any undelivered data, from last_reassem_seq to
	// min(last_block->upper, up_to_seq).
	// Is there such data?
	if ( up_to_seq <= last_reassem_seq ||
	     last_block->upper <= last_reassem_seq )
		return;

	// Skip blocks that are already delivered (but not ACK'ed).
	// Question: shall we instead keep a pointer to the first undelivered
	// block?
	DataBlock* b;
	for ( b = blocks; b && b->upper <= last_reassem_seq; b = b->next )
		;

	ASSERT(b);

	if ( b && b->upper <= up_to_seq )
		{
		if ( DEBUG_tcp_match_undelivered )
			{
			DEBUG_MSG("%.6f matching_undelivered_data: "
			          "last_reassem_seq=%d, up_to_seq=%d, "
			          "first_block=%d-%d\n",
			          network_time, last_reassem_seq, up_to_seq,
			          blocks->seq, blocks->upper);
			}

		contents_processor->Conn()->Weird("matching_undelivered_data");

		for ( ; b && b->upper <= up_to_seq; b = b->next )
			contents_processor->Conn()->Match(Rule::PAYLOAD,
				b->block, b->Size(), false, false,
				contents_processor->IsOrig(), false);
		}
	}

void TCP_Reassembler::RecordToSeq(int start_seq, int stop_seq, BroFile* f)
	{
	DataBlock* b = blocks;
	// Skip over blocks up to the start seq.
	while ( b && b->upper <= start_seq )
		b = b->next;

	if ( ! b )
		return;

	int last_seq = start_seq;
	while ( b && b->upper <= stop_seq )
		{
		if ( b->seq > last_seq )
			RecordGap(last_seq, b->seq, f);

		RecordBlock(b, f);
		last_seq = b->upper;
		b = b->next;
		}

	if ( b )
		// Check for final gap.
		if ( last_seq < stop_seq )
			RecordGap(last_seq, stop_seq, f);
	}

void TCP_Reassembler::RecordBlock(DataBlock* b, BroFile* f)
	{
	unsigned int len = b->Size();
	if ( ! f->Write((const char*)b->block, len) )
		// ### this should really generate an event
		internal_error("contents write failed");
	}

void TCP_Reassembler::RecordGap(int start_seq, int upper_seq, BroFile* f)
	{
	if ( ! f->Write(fmt("\n<<gap %d>>\n", upper_seq - start_seq)) )
		// ### this should really generate an event
		internal_error("contents gap write failed");
	}

void TCP_Reassembler::BlockInserted(DataBlock* start_block)
	{
	if ( start_block->seq > last_reassem_seq ||
	     start_block->upper <= last_reassem_seq )
		return;

	// We've filled a leading hole.  Deliver as much as possible.
	// Note that the new block may include both some old stuff
	// and some new stuff.  AddAndCheck() will have split the
	// new stuff off into its own block(s), but in the following
	// loop we have to take care not to deliver already-delivered
	// data.
	for ( DataBlock* b = start_block; b && b->seq <= last_reassem_seq;
	      b = b->next )
		{
		if ( b->seq == last_reassem_seq )
			{ // New stuff.
			int len = b->Size();
			int seq = last_reassem_seq;

			last_reassem_seq += len;

			if ( record_contents_file )
				RecordBlock(b, record_contents_file);

			contents_processor->DeliverBlock(seq,
							len, b->block);
			if ( rule_matcher )
				contents_processor->Conn()->Match(Rule::PAYLOAD,
					b->block, len, false, false,
					contents_processor->IsOrig(), false);
			}
		}

	TCP_Endpoint* e = contents_processor->Endpoint();

	if ( ! e->peer->HasContents() )
		// Our endpoint's peer doesn't do reassembly and so
		// (presumably) isn't processing acks.  So don't hold
		// the now-delivered data.
		TrimToSeq(last_reassem_seq);

	else if ( e->Size() > max_initial_window && e->NoDataAcked() )
		// We've sent quite a bit of data, yet none of it has
		// been acked.  Presume that we're not seeing the peer's
		// acks (perhaps due to filtering or split routing) and
		// don't hang onto the data further, as we may wind up
		// carrying it all the way until this connection ends.
		TrimToSeq(last_reassem_seq);

	// Note: don't make an EOF check here, because then we'd miss it
	// for FIN packets that don't carry any payload (and thus
	// endpoint->DataSent is not called).  Instead, do the check in
	// TCP_Connection::NextPacket.
	}

void TCP_Reassembler::Overlap(const u_char* b1, const u_char* b2, int n)
	{
	if ( DEBUG_tcp_contents )
		DEBUG_MSG("%.6f TCP contents overlap: %d\n", network_time,  n);

	if ( rexmit_inconsistency &&
	     memcmp((const void*) b1, (const void*) b2, n) &&
	     // The following weeds out keep-alives for which that's all
	     // we've ever seen for the connection.
	     (n > 1 || contents_processor->Endpoint()->peer->HasDoneSomething()) )
		{
		BroString* b1_s = new BroString((const u_char*) b1, n, 0);
		BroString* b2_s = new BroString((const u_char*) b2, n, 0);
		contents_processor->Conn()->Event(rexmit_inconsistency,
					new StringVal(b1_s), new StringVal(b2_s));
		}
	}

IMPLEMENT_SERIAL(TCP_Reassembler, SER_TCP_REASSEMBLER);

bool TCP_Reassembler::DoSerialize(SerialInfo* info) const
	{
	DO_SERIALIZE(SER_TCP_REASSEMBLER, Reassembler);
	SERIALIZE_OPTIONAL(record_contents_file);
	return contents_processor->Serialize(info);
	}

bool TCP_Reassembler::DoUnserialize(UnserialInfo* info)
	{
	DO_UNSERIALIZE(Reassembler);
	UNSERIALIZE_OPTIONAL(record_contents_file, BroFile::Unserialize(info));

	contents_processor = TCP_Contents::Unserialize(info);
	return contents_processor != 0;
	}

TCP_Contents::TCP_Contents(TCP_Endpoint* arg_endp, int arg_stop_on_gap, int arg_punt_on_partial)
	{
	endp = arg_endp;
	endp->AddContentsProcessor(this);
	t_reassem = new TCP_Reassembler(this, endp->src_addr);
	skip_deliveries = 0;
	stop_on_gap = arg_stop_on_gap;
	punt_on_partial = arg_punt_on_partial;
	did_EOF = 0;

	deliver_tcp_contents = 0;
	if ( tcp_contents )
		{
		// Val dst_port_val(ntohs(Conn()->RespPort()), TYPE_PORT);
		PortVal dst_port_val(ntohs(Conn()->RespPort()), TRANSPORT_TCP);
		TableVal* ports = IsOrig() ? 
			tcp_content_delivery_ports_orig : 
			tcp_content_delivery_ports_resp;
		Val* result = ports->Lookup(&dst_port_val);

		if ( (IsOrig() && tcp_content_deliver_all_orig) ||
		     (! IsOrig() && tcp_content_deliver_all_resp) ||
		     (result && result->AsBool()) )
			deliver_tcp_contents = 1;
		}

	if ( punt_on_partial && endp->Conn()->IsPartial() )
		{
		Conn()->SetSkip(1);
		// Conn()->SetRecordPackets(0);	// ###
		}

	seq_to_skip = 0;
	in_delivery = false;
	}

TCP_Contents::~TCP_Contents()
	{
	Unref(t_reassem);
	}

int TCP_Contents::DataSent(double t, int seq, int len,
				const u_char* data)
	{
	int ack = endp->AckSeq() - endp->StartSeq();
	int upper_seq = seq + len;

	if ( DEBUG_tcp_contents )
		{
		DEBUG_MSG("%.6f DataSent: seq=%d upper=%d ack=%d\n",
		          network_time, seq, upper_seq, ack);
		}

	if ( SkipDeliveries() || Conn()->Skipping() )
		return 0;

	if ( seq < ack )
		{
		if ( upper_seq <= ack )
			// We've already delivered this and it's been acked.
			return 0;

		// We've seen an ack for part of this packet, but not the
		// whole thing.  This can happen when, for example, a previous
		// packet held [a, a+b) and this packet holds [a, a+c) for c>b
		// (which some TCP's will do when retransmitting).  Trim the
		// packet to just the unacked data.
		int amount_acked = ack - seq;
		seq += amount_acked;
		data += amount_acked;
		len -= amount_acked;
		}

	t_reassem->NewBlock(t, seq, len, data);

	if ( t_reassem->NumUndeliveredBytes() > max_above_hole_without_any_acks &&
	     Endpoint()->NoDataAcked() )
		{
		Conn()->Weird("above_hole_data_without_any_acks");
		t_reassem->ClearBlocks();
		SetSkipDeliveries(1);
		}

	return 1;
	}

void TCP_Contents::AckReceived(int seq)
	{
	if ( endp->FIN_cnt > 0 && seq >= endp->FIN_seq )
		// TrimToSeq: FIN_seq - 1
		seq = endp->FIN_seq - 1;

	// Note that here SkipDeliveries() and Skipping() are checked after
	// TrimToSeq().

	if ( ! t_reassem->TrimToSeq(seq) &&
	     ! SkipDeliveries() && ! Conn()->Skipping() &&
	     reading_live && endp->state != TCP_PARTIAL && endp->peer->state != TCP_PARTIAL )
		// Only generate these events when reading live; from
		// a trace file, the hole might exist because the
		// file was trimmed while writing it.
		Conn()->Event(ack_above_hole);

	// Check EOF here because t_reassem->LastReassemSeq() may have
	// changed after calling TrimToSeq().
	CheckEOF();
	}

void TCP_Contents::CheckEOF()
	{
	// It is important that the check on whether we have pending data here
	// is consistent with the check in TCP_Connection::ConnnectionClosed().
	//
	// If we choose to call EndpointEOF here because, for example, we
	// are already skipping deliveries, ConnnectionClosed() might decide
	// that there is still DataPending, because it does not check
	// SkipDeliveries(), and the connection will not be closed until
	// timeout, since the did_EOF flag makes sure that EndpointEOF will
	// be called only once.
	//
	// Now both places call TCP_Contents::DataPending(), which checks
	// whether we are skipping deliveries.

	if ( ! did_EOF &&
	     (endp->FIN_cnt > 0 || endp->state == TCP_CLOSED ||
	                           endp->state == TCP_RESET) &&
	     ! DataPending() )
		{
		// We've now delivered all of the data.
		if ( DEBUG_tcp_connection_close )
			{
			DEBUG_MSG("%.6f EOF for %d\n",
			          network_time, endp->IsOrig());
			}

		did_EOF = 1;
		Conn()->EndpointEOF(this);
		}
	}

// DeliverBlock is basically a relay to function Deliver. But unlike
// Deliver, DeliverBlock is not virtual, and this allows us to insert
// operations that apply to all connections using TCP_Contents.

void TCP_Contents::DeliverBlock(int seq, int len, const u_char* data)
	{
	if ( seq + len <= seq_to_skip )
		return;

	if ( seq < seq_to_skip )
		{
		int to_skip = seq_to_skip - seq;
		len -= to_skip;
		data += to_skip;
		seq = seq_to_skip;
		}

	if ( deliver_tcp_contents )
		{
		val_list* vl = new val_list();
		vl->append(Conn()->BuildConnVal());
		vl->append(new Val(IsOrig(), TYPE_BOOL));
		vl->append(new Val(seq, TYPE_COUNT));
		vl->append(new StringVal(len, (const char*) data));

		Conn()->ConnectionEvent(tcp_contents, vl);
		}

	// Q. Can we say this because it is already checked in DataSent()?
	// ASSERT(!Conn()->Skipping() && !SkipDeliveries());
	//
	// A. No, because TrimToSeq() can deliver some blocks after
	// skipping the undelivered.

	if ( Conn()->Skipping() || SkipDeliveries() )
 		return;

 	if ( punt_on_partial &&
 	     (endp->state == TCP_PARTIAL || endp->peer->state == TCP_PARTIAL) )
 		{
 		Conn()->SetSkip(1);
 		// Conn()->SetRecordPackets(0);	// ###
 		return;
 		}

	in_delivery = true;
	Deliver(seq, len, data);
	in_delivery = false;

	if ( seq + len < seq_to_skip )
		DoSkipToSeq(seq_to_skip);
	}

void TCP_Contents::Deliver(int seq, int len, const u_char* data)
	{
	Conn()->Deliver(endp, seq, len, data);
	}

void TCP_Contents::Undelivered(int seq, int len)
	{
	if ( DEBUG_tcp_contents )
		{
		DEBUG_MSG("%.6f Undelivered: seq=%d, len=%d, "
		          "skip_deliveries=%d\n",
		          network_time, seq, len, skip_deliveries);
		}

	if ( ! Conn()->Skipping() && ! SkipDeliveries() )
		Conn()->Undelivered(endp, seq, len);

	if ( stop_on_gap )
		{
		// No more data will be sent to the reassembler via
		// DataSent, and thus no data will be delivered. Note
		// that it is safe to stop right here because all the data
		// before seq that exists in the trace must have already
		// appeared.
		// Also note that SetSkipDeliveries(1) takes place after 
		// Conn()->Undelivered(), so that conn will still see
		// the first Undelivered().

		SetSkipDeliveries(1);
		}
	}

void TCP_Contents::DoSkipToSeq(int seq)
	{
	t_reassem->SkipToSeq(seq);
	}

void TCP_Contents::SkipToSeq(int seq)
	{
	if ( seq > seq_to_skip )
		{
		seq_to_skip = seq;
		if ( ! in_delivery )
			DoSkipToSeq(seq);
		}
	}

void TCP_Contents::SizeBufferedData(int& waiting_on_hole, int& waiting_on_ack)
	{
	t_reassem->SizeBufferedData(waiting_on_hole, waiting_on_ack);
	}

int TCP_Contents::DataPending() const
	{
	// If we are skipping deliveries, the reassembler will not get called
	// in DataSent(), and DataSeq() will not be updated.
	if ( Conn()->Skipping() || SkipDeliveries() )
		return 0;

	uint32 delivered_seq = Endpoint()->StartSeq() + DataSeq();

	// Q. Can we say that?
	// ASSERT(delivered_seq <= Endpoint()->LastSeq());
	// 
	// A. Yes, but only if we express it with 64-bit comparison
	// to handle sequence wrapping around

	// We've delivered everything if we're up to the penultimate
	// sequence number (since a FIN consumes an octet in the
	// sequence space), or right at it (because a RST does not).
	if ( delivered_seq != Endpoint()->LastSeq() - 1 &&
	     delivered_seq != Endpoint()->LastSeq() )
		return 1;

	// If we've sent RST, then we can't send ACKs any more.
	if ( Endpoint()->state != TCP_RESET &&
	     Endpoint()->peer->HasUndeliveredData() )
		return 1;

	return 0;
	}

void TCP_Contents::SetContentsFile(BroFile* f)
	{
	t_reassem->SetContentsFile(f);
	}

IMPLEMENT_SERIAL(TCP_Contents, SER_TCP_CONTENTS);

bool TCP_Contents::Serialize(SerialInfo* info) const
	{
	return SerialObj::Serialize(info);
	}

TCP_Contents* TCP_Contents::Unserialize(UnserialInfo* info)
	{
	return (TCP_Contents*) SerialObj::Unserialize(info, SER_TCP_CONTENTS);
	}

bool TCP_Contents::DoSerialize(SerialInfo* info) const
	{
	DO_SERIALIZE(SER_TCP_CONTENTS, SerialObj);

	// FIXME: We don't serialize the reassembler. Would this make sense?
	return endp->Serialize(info) &&
		t_reassem->Serialize(info) &&
		SERIALIZE_BIT(skip_deliveries) &&
		SERIALIZE_BIT(stop_on_gap) &&
		SERIALIZE_BIT(did_EOF) &&
		SERIALIZE_BIT(punt_on_partial);
	}

bool TCP_Contents::DoUnserialize(UnserialInfo* info)
	{
	DO_UNSERIALIZE(SerialObj);
	endp = TCP_Endpoint::Unserialize(info);
	if ( ! endp )
		return false;

	t_reassem =(TCP_Reassembler*) Reassembler::Unserialize(info);
	if ( ! t_reassem )
		return false;

	UNSERIALIZE_BIT(skip_deliveries);
	UNSERIALIZE_BIT(stop_on_gap);
	UNSERIALIZE_BIT(did_EOF);
	UNSERIALIZE_BIT(punt_on_partial);

	return true;
	}

TCP_ContentLine::TCP_ContentLine(TCP_Endpoint* arg_endp,
			int is_NUL_sensitive, int skip_partial,
		        int arg_CRLF_as_EOL)
: TCP_Contents(arg_endp, 0, skip_partial)
	{
	flag_NULs = is_NUL_sensitive;
	CR_LF_as_EOL = arg_CRLF_as_EOL;
	line_limit = 1024;
	buf = 0;
	Init(0);
	seq_delivered_in_lines = 0;
	skip_pending = 0;
	}

void TCP_ContentLine::Init(int size)
	{
	if ( size < 128 )
		size = 128;

	char* b = new char[size];

	if ( buf )
		{
		if ( offset > 0 )
			memcpy(b, buf, offset);
		delete [] buf;
		}
	else
		{
		offset = 0;
		last_char = 0;
		}

	buf = b;
	buf_len = size;
	}

TCP_ContentLine::~TCP_ContentLine()
	{
	delete [] buf;
	}

void TCP_ContentLine::Deliver(int seq, int len, const u_char* data)
	{
	if ( len <= 0 || SkipDeliveries() )
		return;

	if ( buf && len + offset >= buf_len )
		{ // Make sure we have enough room to accommodate the new stuff.
		int old_buf_len = buf_len;
		buf_len = ((offset + len) * 3) / 2 + 1;

		char* tmp = new char[buf_len];
		for ( int i = 0; i < old_buf_len; ++i )
			tmp[i] = buf[i];

		delete [] buf;
		buf = tmp; 

		if ( ! buf )
			internal_error("out of memory delivering endpoint line");
		}

	DoDeliver(seq, len, data);
	}

const u_char* TCP_ContentLine::NextNewLine(int len, const u_char* data, int& next_pos)
	{
	const u_char* end_of_data = data + len;
	for ( const u_char* p = data; p < end_of_data; ++p )
		if ( *p == '\n' || *p == '\r' )
			{
			next_pos = p + 1 - data;	// where next line begins
			last_char = *p;
			return p;
			}
	next_pos = 0;
	return 0;
	}


void TCP_ContentLine::DoDeliver(int seq, int len, const u_char* data)
	{
	seq_delivered_in_lines = seq;
	while ( len > 0 && ! SkipDeliveries() )
		{
		if ( (CR_LF_as_EOL & CR_as_EOL) &&
		     last_char == '\r' && *data == '\n' )
			{
			// CR is already considered as EOL.
			// Compress CRLF to just one line termination.
			last_char = *data;
			--len; ++data; ++seq;
			++seq_delivered_in_lines;
			}

		if ( skip_pending > 0 )
			SkipBytes(skip_pending);

		// Note that the skipping must take place *after*
		// the CR/LF check above, so that the '\n' of the
		// previous line is skipped first.
		if ( seq < seq_to_skip )
			{
			// Skip rest of the data and return
			int skip_len = seq_to_skip - seq;
			if ( skip_len > len )
				skip_len = len;

			Undelivered(seq, skip_len);

			len -= skip_len; data += skip_len;
			seq += skip_len;
			seq_delivered_in_lines += skip_len;
			}

		if ( len <= 0 )
			break;

		int n = DoDeliverOnce(seq, len, data);
		len -= n;
		data += n;
		seq += n;
		}
	}

int TCP_ContentLine::DoDeliverOnce(int seq, int len, const u_char* data)
	{
	const u_char* data_start = data;

	// ### Note, excessive_line checking no longer done.

	// ### This code should be correspondingly updated in TCP_NVT
	// ### and Rlogin.
	//        fprintf(stderr, "Data delivered (len): '%.*s' (%d)\n", len, data, len);

	if ( len <= 0 )
		return 0;

	for ( ; len > 0; --len, ++data )
		{
		if ( offset >= buf_len )
			Init(buf_len * 2);

		int c = data[0];

#define EMIT_LINE \
	{ \
	buf[offset] = '\0'; \
	int seq_len = data + 1 - data_start; \
	seq_delivered_in_lines = seq + seq_len; \
	last_char = c; \
	Conn()->NewLine(this, offset, buf); \
	offset = 0; \
	return seq_len; \
	}

		switch ( c ) {
		case '\r':
			// Look ahead for '\n'.
			if ( len > 1 && data[1] == '\n' )
				{
				--len; ++data;
				last_char = c;
				c = data[0];
				EMIT_LINE
				}
			else if ( CR_LF_as_EOL & CR_as_EOL )
				EMIT_LINE
			else
				buf[offset++] = c;
			break;

		case '\n':
			if ( last_char == '\r' )
				{
				--offset; // remove '\r'
				EMIT_LINE
				}
			else if ( CR_LF_as_EOL & LF_as_EOL )
				EMIT_LINE
			else
				{
				if ( Conn()->FlagEvent(SINGULAR_LF) )
					Conn()->Weird("line_terminated_with_single_LF");
				buf[offset++] = c;
				}
			break;

		case '\0':
			if ( flag_NULs )
				CheckNUL();
			else
				buf[offset++] = c;
			break;

		default:
			buf[offset++] = c;
			break;
		}

		if ( last_char == '\r' )
			if ( Conn()->FlagEvent(SINGULAR_CR) )
				Conn()->Weird("line_terminated_with_single_CR");

		last_char = c;
		}

	return data - data_start;
	}

void TCP_ContentLine::CheckNUL()
	{
	// If this is the first byte seen on this connection,
	// and if the connection's state is PARTIAL, then we've
	// intercepted a keep-alive, and shouldn't complain
	// about it.  Note that for PARTIAL connections, the
	// starting sequence number is adjusted as though there
	// had been an initial SYN, so we check for whether
	// the connection has at most two bytes so far.
	if ( endp->state == TCP_PARTIAL &&
	     endp->LastSeq() - endp->StartSeq() <= 2 )
		; // Ignore it.
	else
		{
		if ( Conn()->FlagEvent(NUL_IN_LINE) )
			Conn()->Weird("NUL_in_line");
		flag_NULs = 0;
		}
	}

void TCP_ContentLine::ExcessiveLine(const u_char* /* data */, int /* len */)
	{
	Conn()->Event(excessive_line);
	line_limit = 0;
	}

void TCP_ContentLine::SkipBytesAfterThisLine(int length)
	{
	// This is a little complicated because Bro has to handle
	// both CR and CRLF as a line break. When a line is delivered,
	// it's possible that only a CR is seen, and we may not know
	// if an LF is following until we see the next packet. If an
	// LF follows, we should start skipping bytes *after* the LF.
	// So we keep the skip as 'pending' until we see the next
	// character in DoDeliver().

	if ( last_char == '\r' )
		skip_pending = length;
	else
		SkipBytes(length);
	}

void TCP_ContentLine::SkipBytes(int length)
	{
	skip_pending = 0;
	SkipToSeq(SeqDelivered() + length);
	}

IMPLEMENT_SERIAL(TCP_ContentLine, SER_TCP_CONTENT_LINE);

bool TCP_ContentLine::DoSerialize(SerialInfo* info) const
	{
	DO_SERIALIZE(SER_TCP_CONTENT_LINE, TCP_Contents);

	return SERIALIZE_BIT(flag_NULs) &&
		SERIALIZE_BIT(line_limit) &&
		SERIALIZE((unsigned int)CR_LF_as_EOL) &&
	SERIALIZE_STR(buf, offset) &&
		SERIALIZE(offset) &&
		SERIALIZE(buf_len) &&
		SERIALIZE(last_char);
	}

bool TCP_ContentLine::DoUnserialize(UnserialInfo* info)
	{
	DO_UNSERIALIZE(TCP_Contents);

	UNSERIALIZE_BIT(flag_NULs);
	UNSERIALIZE_BIT(line_limit);

	unsigned int tmp;
	if ( ! UNSERIALIZE(&tmp) )
		return false;

	CR_LF_as_EOL = tmp;

	const char* tmpbuf;
	int tmplen;

	if ( ! (UNSERIALIZE_STR(&tmpbuf, &tmplen) &&
		UNSERIALIZE(&offset) &&
		UNSERIALIZE(&buf_len) &&
		UNSERIALIZE(&last_char)) )
		return false;

	// If our input is correct, tmplen is <= buf_len. Check this.
	if ( tmplen > buf_len )
		{
		info->s->Error("incorrect buffer size in TCP_ContentLine");
		return false;
		}

	buf = new char[buf_len];
	memcpy(buf, tmpbuf, tmplen);

	return true;
	}
